from mock import Mock
from tests.helpers.testbase import TestBase
from cerami.request.mixins import Pageable
from cerami.request.search_attribute import DictAttribute

class TestRequest(Pageable):
    pass

class TestPageable(TestBase):
    def setUp(self):
        self.request = TestRequest()

    def test_start_key(self):
        """it adds the dict to as ExclusiveStartKey"""
        self.request.add_attribute = Mock()
        self.request.start_key({'fakeDict': True})
        self.request.add_attribute.assert_called_with(
            DictAttribute,
            'ExclusiveStartKey',
            {'fakeDict': True})



from mock import Mock, patch
from datetime import datetime, timezone
from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype.expression import (
    EqualityExpression,
    InExpression,
    ListAppendExpression,
    ArithmeticExpression)
from cerami.datatype.mapper import (
    ByteMapper,
    StringMapper,
    ModelMapper,
    DictMapper,
    ListMapper,
    IntegerMapper,
    DecimalMapper,
    DatetimeMapper,
    SetMapper)
from cerami.datatype import (
    DynamoDataType,
    BaseString,
    BaseNumber,
    ByteBuffer,
    String,
    Datetime,
    Map,
    ModelMap,
    List,
    Set)

class TestBaseString(TestBase):
    def setUp(self):
        super(TestBaseString, self).setUp()
        self.dt = BaseString(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "S"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, StringMapper)

class TestBaseNumber(TestBase):
    def setUp(self):
        super(TestBaseNumber, self).setUp()
        self.dt = BaseNumber(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "N"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, IntegerMapper)

    def test_add(self):
        """it returns an ArithmeticExpression"""
        res = self.dt.add(5)
        assert isinstance(res, ArithmeticExpression)
        assert res.expression == '+'

    def test_subtract(self):
        """it returns an ArithmeticExpression"""
        res = self.dt.subtract(5)
        assert isinstance(res, ArithmeticExpression)
        assert res.expression == '-'

class TestString(TestBase):
    def setUp(self):
        super(TestString, self).setUp()
        self.dt = String(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "S"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, StringMapper)

class TestByteBuffer(TestBase):
    def setUp(self):
        super(TestByteBuffer, self).setUp()
        self.dt = ByteBuffer(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "B"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, ByteMapper)

class TestDatetime(TestBase):
    def setUp(self):
        super(TestDatetime, self).setUp()
        self.dt = Datetime(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "S"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, DatetimeMapper)

class TestMap(TestBase):
    def setUp(self):
        super(TestMap, self).setUp()
        self.mocked_guesser = Mock()
        self.dt = Map(
            map_guesser=self.mocked_guesser,
            parse_guesser=self.mocked_guesser)

    def test__init__(self):
        """it initializes the guesser"""
        assert self.dt.map_guesser == self.mocked_guesser
        assert self.dt.parse_guesser == self.mocked_guesser

    def test_condition_type(self):
        assert self.dt.condition_type == "M"

    def test_mapper(self):
        """it is an instance of DictMapper"""
        assert isinstance(self.dt.mapper, DictMapper)

class TestList(TestBase):
    def setUp(self):
        super(TestList, self).setUp()
        self.mocked_guesser = Mock()
        self.dt = List(
            column_name="test",
            map_guesser=self.mocked_guesser,
            parse_guesser=self.mocked_guesser)

    def test__init__(self):
        """it sets guesser"""
        assert self.dt.map_guesser == self.mocked_guesser
        assert self.dt.parse_guesser == self.mocked_guesser

    def test_property(self):
        assert self.dt.condition_type == "L"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, ListMapper)

    def test_append(self):
        """it should return a ListAppendExpression"""
        res = self.dt.append(['Test'])
        assert isinstance(res, ListAppendExpression)
        assert res.value == ['Test']

    def test_append_scalar(self):
        """it converts the value to a list when it isnt one"""
        res = self.dt.append('Test')
        assert isinstance(res, ListAppendExpression)
        assert res.value == ['Test']

    def test_index(self):
        """it returns a new instance of the datatype passed with the column_name
        it sets _index on the returned datatype instance"""
        res = self.dt.index(0, String())
        assert isinstance(res, String)
        assert res._index == 0

    def test_index_throws_ValueEreror(self):
        """it throws a ValueError when the datatype is not an instance
        This is to prevent people from accidentially passing the class in and breaking
        everything
        """
        self.assertRaises(ValueError, self.dt.index, 0, String)

class TestModelMap(TestBase):
    class TestModel(Model):
        __tablename__ = "test"
        test1 = String()
        test2 = String()

    def setUp(self):
        super(TestModelMap, self).setUp()
        self.dt = ModelMap(self.TestModel, column_name="test")

    def test__init__(self):
        """it should setattr for each column in the model_cls"""
        for column in self.TestModel._columns:
            assert hasattr(self.dt, column.column_name)
            assert isinstance(getattr(self.dt, column.column_name),
                              type(column))

    def test_condition_type(self):
        assert self.dt.condition_type == "M"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, ModelMapper)

    def set_column_name(self):
        """it updates the column_name and all nested column names"""
        new_name = "test2"
        self.dt.set_column_name(new_name)
        assert self.dt.column_name == new_name
        for column in TestModel._columns:
            nested_column = getattr(self.dt, column.column_name)
            new_nested_name = new_name + "." + column.column_name
            assert nested_column.column_name == new_nested_name

class TestSet(TestBase):
    def setUp(self):
        super(TestSet, self).setUp()
        self.dt = Set(String(), column_name="test")

    def test_has_a_datatype(self):
        assert isinstance(self.dt.datatype, String)

    def test_condition_type(self):
        """it adds an S to the datatype's condition_type"""
        assert self.dt.condition_type == "SS"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, SetMapper)

class TestDynamoDataType(TestBase):
    def setUp(self):
        super(TestDynamoDataType, self).setUp()
        self.dt = DynamoDataType(column_name="test")
        self.dt.mapper = StringMapper(self.dt)
        self.dt.condition_type = "S"

    def test__eq__(self):
        res = self.dt == 1
        assert isinstance(res, EqualityExpression)
        assert res.expression == "="

    def test__ne__(self):
        res = self.dt != 1
        assert isinstance(res, EqualityExpression)
        assert res.expression == "<>"

    def test__gt__(self):
        res = self.dt > 1
        assert isinstance(res, EqualityExpression)
        assert res.expression == ">"

    def test__ge__(self):
        res = self.dt >= 1
        assert isinstance(res, EqualityExpression)
        assert res.expression == ">="

    def test__lt__(self):
        res = self.dt < 1
        assert isinstance(res, EqualityExpression)
        assert res.expression == "<"

    def test__le__(self):
        res = self.dt <= 1
        assert isinstance(res, EqualityExpression)
        assert res.expression == "<="

    def test_in_(self):
        res = self.dt.in_('1','2','3')
        assert isinstance(res, InExpression)



from decimal import Decimal
from .base_datatype_mapper import BaseDatatypeMapper

class DecimalMapper(BaseDatatypeMapper):
    """A Mapper class for converting decimal number fields into DynaomDB dictionaries

    For example::

        mapper = DecimalMapper(Number())
        mapper.map(30)
        {'N': '30'}

        mapper.map(30.69213)
        {'N': '30.69213'}

        mapper.reconstruct({'N': '30'})
        Decimal('30')

        mapper.reconstruct({'N': '30.69213'})
        Decimal('30.69213')
    """
    def resolve(self, value):
        """convert the number into a decimal string"""
        return str(Decimal(str(value)))

    def parse(self, value):
        """convert the value back into a Decimal"""
        return Decimal(value)



from ..search_attribute import DictAttribute

class Pageable(object):
    """A mixin to add the start_key method

    Pageable requests can be paginated by passing in an ExclusiveStartKey attribute.
    """

    def start_key(self, key):
        """return a new Request setup with the ExclusiveStartKey attribute

        Parameters:
            key: A dict representing the exclusive start key. This key is _excluded_
                from the response. Typically, this key is auto-generated from a previous
                response of a request.


        For example::

            request = Person.scan \\
                .limit(1) \\
                .start_key({"email": {"S": "test@test.com"}})

            request.build()
            {
                "TableName": "people",
                "Limit": 1,
                "ExclusiveStartKey": {
                    "email": {
                        "S": "test@test.com",
                    },
                },
            }

            # you can use the responses exclusive_start_key to continually paginate
            response = request.execute()
            next_response = request.start_key(response.last_evaludated_key).execute()
        """
        self.add_attribute(
            DictAttribute,
            'ExclusiveStartKey',
            key)
        return self



from .base_datatype import DynamoDataType
from .expression import (
    ContainsExpression,
    BeginsWithExpression)
from .mapper import StringMapper

class BaseString(DynamoDataType):
    """A Base class for all String datatypes"""

    def __init__(self, mapper_cls=StringMapper, default=None, column_name=""):
        """constructor for the BaseString

        Parameters:
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
            mapper_cls: A mapper class to manipulate data to/from dynamodb.
                Defaults to the StringMapper
        """
        super(BaseString, self).__init__(
            mapper_cls=mapper_cls,
            condition_type="S",
            default=default,
            column_name=column_name)

    def begins_with(self, value):
        """Build a BeginsWithExpression

        Can be used in Filters or KeyConditionExpressions to create a begins_with
        expression.

        Parameters:
            value: a substring to check if the column begins with

        Returns:
            A BeginsWithExpression

        For example::

            Person.scan.filter(Person.name.begins_with("Mo"))
        """
        return BeginsWithExpression(self, value)

    def contains(self, value):
        """Build a ContainsExpression

        Can be used in Filters only, cannot be part of a KeyConditionExpression

        Parameters:
            value: the value to filter upon

        Returns:
            A ContainsExpression

        For example::

            Person.scan.filter(Person.name.contains("om"))
        """
        return ContainsExpression(self, value)



from mock import Mock, patch, call
from tests.helpers.testbase import TestBase
from cerami.response import SaveResponse
from cerami.request import UpdateRequest
from cerami.request.mixins import Keyable, Returnable
from cerami.datatype import String
from cerami.datatype.expression import EqualityExpression
from cerami.request.search_attribute import (
    UpdateExpressionAttribute,
    UpdateAction,
    DictAttribute)

class TestUpdateRequest(TestBase):
    def setUp(self):
        self.mocked_client = Mock()
        self.request  = UpdateRequest(
            tablename="test",
            client=self.mocked_client)

    def test_is_keyable(self):
        """it is keyable"""
        assert isinstance(self.request, Keyable)

    def test_is_returnable(self):
        """it is returnable"""
        assert isinstance(self.request, Returnable)

    def test_set(self):
        """it adds the SET attribute"""
        expression = EqualityExpression('=', String(), 'test')
        self.request.update_expression = Mock()
        self.request.set(expression)
        self.request.update_expression.assert_called_with('SET', expression)



from tests.helpers.testbase import TestBase
from cerami.datatype import Number
from cerami.datatype.mapper import DecimalMapper
from decimal import Decimal

class TestDecimalMapper(TestBase):
    def setUp(self):
        self.dt = Number()
        self.mapper = DecimalMapper(self.dt)

    def test_resolve(self):
        """it returns the value as a string"""
        assert self.mapper.resolve(1) == '1'
        assert self.mapper.resolve(123.456) == '123.456'

    def test_parse(self):
        """it returns the value as a decimal"""
        assert self.mapper.parse('1') == Decimal('1')
        assert self.mapper.parse('123.456') == Decimal('123.456')
